# -*- coding: UTF-8 -*-
"""
Unit tests for the ``iiqtools.iiq_tar_to_zip`` module's business logic
"""
import os
import unittest
from mock import patch, MagicMock

from iiqtools import iiq_tar_to_zip


class TestBufferedZipFile(unittest.TestCase):
    """A suite of tests for the BufferedZipFile class"""

    def setUp(self):
        """Runs before every test case"""
        self.filepath = '/tmp/foo.zip'
        self.zipfile = iiq_tar_to_zip.BufferedZipFile('/tmp/foo.zip', mode='w')
        self.zipfile.fp = MagicMock()
        self.zipfile._writecheck = MagicMock()

        self.fake_file = MagicMock()
        self.fake_file.read.side_effect = ['asdf', '']

    def tearDown(self):
        """Runs after every tests case"""
        os.remove(self.filepath)

    @patch.object(iiq_tar_to_zip, 'struct')
    @patch.object(iiq_tar_to_zip, 'binascii')
    def test_basic(self, fake_binascii, fake_struct):
        """BufferedZipFile - writebuffered is callable"""
        self.zipfile.writebuffered(filename='foo', file_handle=self.fake_file)


class TestCheckTar(unittest.TestCase):
    """A suite of tests for the check_tar function"""

    @patch.object(iiq_tar_to_zip.os.path, 'isfile')
    @patch.object(iiq_tar_to_zip.tarfile, 'is_tarfile')
    def test_valid_tar(self, fake_is_tarfile, fake_isfile):
        """The supplied tar value is returned when it's a valid tar file"""
        fake_is_tarfile.return_value = True
        fake_isfile.return_value = True

        sent = 'insightiq_export_1234567890.tar.gz'
        returned = iiq_tar_to_zip.check_tar(sent)

        self.assertEqual(sent, returned)

    def test_not_a_file(self):
        """argparse.ArgumentTypeError is raised if the supplied tar file doesn't exist"""
        sent = 'insightiq_export_1234567890.tar.gz'
        self.assertRaises(iiq_tar_to_zip.argparse.ArgumentTypeError, iiq_tar_to_zip.check_tar, sent)

    @patch.object(iiq_tar_to_zip.os.path, 'isfile')
    @patch.object(iiq_tar_to_zip.tarfile, 'is_tarfile')
    def test_not_a_tar(self, fake_is_tarfile, fake_isfile):
        """argparse.ArgumentTypeError is raised if the supplied file isn't a tar file"""
        fake_isfile.return_value = True
        fake_is_tarfile.return_value = False
        sent = 'insightiq_export_1234567890.tar.gz'
        self.assertRaises(iiq_tar_to_zip.argparse.ArgumentTypeError, iiq_tar_to_zip.check_tar, sent)

    @patch.object(iiq_tar_to_zip.os.path, 'isfile')
    @patch.object(iiq_tar_to_zip.tarfile, 'is_tarfile')
    def test_bad_file_name(self, fake_is_tarfile, fake_isfile):
        """argparse.ArgumentTypeError is raised if the tar file doesn't adhere to the expected naming convention"""
        fake_is_tarfile.return_value = True
        fake_isfile.return_value = True

        sent = 'insightiq_export_1.tar.gz'
        self.assertRaises(iiq_tar_to_zip.argparse.ArgumentTypeError, iiq_tar_to_zip.check_tar, sent)


class TestParseCli(unittest.TestCase):
    """A suite of tests for the parse_cli function"""

    # Mocking away stderr to avoid spam in output while running tests
    @patch.object(iiq_tar_to_zip.argparse._sys, 'stderr')
    def test_no_args(self, fake_stderr):
        """When no cli args are given, SystemExit is raised"""
        self.assertRaises(SystemExit, iiq_tar_to_zip.parse_cli, [])

    @patch.object(iiq_tar_to_zip.argparse._sys, 'stderr')
    def test_missing_required(self, fake_stderr):
        """When missing required args, SystemExit is raised"""
        self.assertRaises(SystemExit, iiq_tar_to_zip.parse_cli, ['-o', '/tmp'])

    @patch.object(iiq_tar_to_zip, 'check_tar')
    @patch.object(iiq_tar_to_zip, 'check_path')
    def test_returns_namespace(self, fake_check_path, fake_check_tar):
        fake_check_path.return_value = '/tmp'
        fake_check_tar.return_value = 'insightiq_export_1234.tar.gz'

        args = iiq_tar_to_zip.parse_cli(['--source-tar', 'insightiq_export_1234.tar.gz', '--output-dir', '/tmp'])

        self.assertTrue(isinstance(args, iiq_tar_to_zip.argparse.Namespace))


class TestGetTimestampFromExport(unittest.TestCase):
    """A suite of tests for the get_timestamp_from_export function"""

    def test_absolute_path(self):
        """get_timestamp_from_export works if source_tar is a file path"""
        source_tar = '/datastore/insightiq_export_1234.tar.gz'

        actual = iiq_tar_to_zip.get_timestamp_from_export(source_tar)
        expected = '1234'

        self.assertEqual(actual, expected)


class TestJoinname(unittest.TestCase):
    """A suite of tests for the joinname function"""

    def test_relative_path(self):
        """joinname - works with relative paths"""
        the_dir = 'insightiq_export_1234567890'
        the_file = '../somefile'

        actual = iiq_tar_to_zip.joinname(the_dir, the_file)
        expected = 'insightiq_export_1234567890/somefile'

        self.assertEqual(actual, expected)

    def test_absolute_path(self):
        """joinname - works with absolute paths"""
        the_dir = 'insightiq_export_1234567890'
        the_file = '/datastore/somefile'

        actual = iiq_tar_to_zip.joinname(the_dir, the_file)
        expected = 'insightiq_export_1234567890/somefile'

        self.assertEqual(actual, expected)


class TestMain(unittest.TestCase):
    """A suite of tests for the main function"""

    def setUp(self):
        """Runs before every test case"""
        # So we don't have to patch the same stuff in every test case
        self.patches = {'fake_logger' : patch('iiqtools.iiq_tar_to_zip.get_logger'),
                        'fake_BufferedZipFile' : patch('iiqtools.iiq_tar_to_zip.BufferedZipFile'),
                        'fake_parse_cli' : patch('iiqtools.iiq_tar_to_zip.parse_cli'),
                        'fake_tarfile' : patch('iiqtools.iiq_tar_to_zip.tarfile'),
                        'fake_os_remove' : patch('iiqtools.iiq_tar_to_zip.os.remove'),
                       }
        for patch_name, the_patch in self.patches.items():
            patched_obj = the_patch.start()
            setattr(self, patch_name, patched_obj)

        self.fake_parse_cli.source_tar = 'insightiq_export_1234567890.tar.gz'
        self.fake_parse_cli.output_dir = '/tmp'
        fake_file1 = MagicMock()
        fake_file1.name = 'foo'
        fake_file2 = MagicMock()
        fake_file2.name = 'bar'
        self.fake_tarfile.open.return_value.getmembers.return_value = [fake_file1, fake_file2]

    def tearDown(self):
        """Runs after every test case"""
        for the_patch in self.patches.values():
            the_patch.stop()

    def test_basic(self):
        """The main function returns zero when there are no errors"""
        exit_code = iiq_tar_to_zip.main(['-s', 'insightiq_export_1234567890.tar.gz', '-o', '/tmp'])
        self.assertEqual(exit_code, 0)

    def test_zip_creation_failure(self):
        """The main function returns 1 when it cannot create the new zip file"""
        self.fake_BufferedZipFile.side_effect = [IOError()]
        exit_code = iiq_tar_to_zip.main(['-s', 'insightiq_export_1234567890.tar.gz', '-o', '/tmp'])

        self.assertEqual(exit_code, 1)

    def test_mid_write_failure(self):
        """
        The main function returns the error code of the exception if it fails
        while writing to the new zip file
        """
        self.fake_BufferedZipFile.return_value.writebuffered.side_effect = [IOError(13, 'testing', 'some_file')]
        exit_code = iiq_tar_to_zip.main(['-s', 'insightiq_export_1234567890.tar.gz', '-o', '/tmp'])

        self.assertEqual(exit_code, 13)
        self.assertTrue(self.fake_os_remove.called)


if __name__ == '__main__':
    unittest.main()



# -*- UTF-8 -*-
"""
Centralized location for all custom Exceptions
"""

class CliError(Exception):
    """Raised when an CLI command has a non-zero exit code.

    :param command: The CLI command that was ran
    :type command: String

    :param stdout: The output from the standard out stream
    :type stdout: String

    :param stderr: The output from the standard error stream
    :type stderr: String

    :param exit_code: The exit/return code from the command
    :type exit_codee: Integer
    """
    def __init__(self, command, stdout, stderr, exit_code, message='Command Failure'):
        msg = '%s: %s' % (message, command)
        super(CliError, self).__init__(msg)
        self.command = command
        self.stdout = stdout
        self.stderr = stderr
        self.exit_code = exit_code


class DatabaseError(Exception):
    """Raised when an error occurs when interacting with the InsightIQ database

    :attribute pgcode: The error code used by PostgreSQL. https://www.postgresql.org/docs/9.3/static/errcodes-appendix.html
    :attribute message: The error message
    """
    def __init__(self, message, pgcode):
        super(DatabaseError, self).__init__(message)
        self.pgcode = pgcode



# -*- coding: UTF-8 -*-

# Well, you might be wondering what the hell we're doing in this file.
# The short answer is hacking on sys.path so Python can find the few packages
# in InsightIQ used by IIQTools. Yes, I know, not the most elegant solution; simply
# installing those dependencies would be the "right way" to do this, but that's not
# possible here. InsightIQ is closed source. In addition, installing everything
# except InsightIQ (to cut down on the path hacking) doesn't work for all our
# users; some have no access to the Internet, thus they'd have to manually install
# all these packages.
# The path hacking is happening here because this __init__ will alway be evaluated
# when importing anything from IIQTools. The only other option would be to add
# the path hacks to *every* script, which is a far worse idea. The great news is
# that unless you've stumbled upon this file, you'd never have known about it;
# everything should "just work" as far as importing packages goes.
import sys

IIQ_PATHS = ['/usr/share/isilon/lib/python2.6/site-packages/argparse-1.2.1-py2.6.egg',
             '/usr/share/isilon/lib/python2.6/site-packages/python_cjson-1.0.5-py2.6-linux-x86_64.egg',
             '/usr/share/isilon/lib/python2.6/site-packages/Paste-1.7.5.1-py2.6.egg',
             '/usr/share/isilon/lib/python2.6/site-packages/PasteDeploy-1.3.3-py2.6.egg',
             '/usr/share/isilon/lib/python2.6/site-packages/Pylons-1.0-py2.6.egg',
             '/usr/share/isilon/lib/python2.6/site-packages/WebHelpers-0.6.4-py2.6.egg',
             '/usr/share/isilon/lib/python2.6/site-packages/dnspython-1.8.0-py2.6.egg',
             '/usr/share/isilon/lib/python2.6/site-packages/psycopg2-2.5.4-py2.6-linux-x86_64.egg',
             '/usr/share/isilon/lib/python2.6/site-packages/python_cjson-1.0.5-py2.6-linux-x86_64.egg',
             '/usr/share/isilon/lib/python2.6/site-packages/pytz-2015.2-py2.6.egg',
             '/usr/share/isilon/lib/python2.6/site-packages/configobj-4.7.2-py2.6.egg',
]

sys.path.insert(1, '/usr/share/isilon/lib/python2.6/site-packages')
sys.path.insert(1, '/usr/share/isilon/lib/python2.7/site-packages')
sys.path += IIQ_PATHS



# -*- coding: UTF-8 -*-
"""
This module contains miscellaneous utility functions
"""
import os.path
import argparse
from sys import stderr


def printerr(message):
    """Just like print(), but outputs to stderr.

    :Returns: None

    :param message: The thing to write to stderr
    :type message: PyObject
    """
    msg = '%s\n' % message # print() auto-inserts a line return too
    stderr.write(msg)
    stderr.flush()


def check_path(cli_value):
    """Validate that the supplied path is an actual file system directory.

    This function is intended to be used with the argparse lib as an
    `argument type <https://docs.python.org/2.7/library/argparse.html#type>`_.

    :Raises: argparse.ArgumentTypeError

    :Returns: String

    :param cli_value: The value supplied by the end user
    :type cli_value: String
    """
    if not os.path.isdir(cli_value):
        msg = 'Supplied value is not a directory, %s' % (cli_value)
        raise argparse.ArgumentTypeError(msg)
    else:
        return cli_value



# -*- coding: UTF-8 -*-
"""
This module contains all the business logic for patching InsightIQ.
"""
from __future__ import print_function

import os
import sys
import glob
import shutil
import tarfile
import hashlib
import getpass
import argparse
from StringIO import StringIO
from collections import namedtuple

from configobj import ConfigObj

from iiqtools.utils import shell, versions
from iiqtools.utils.logger import get_logger


#########################################
# Data structures and related functions #
#########################################


_PatchContents = namedtuple("PatchContents", "readme meta_ini patched_files")
# So we can have a nice docstring for the namedtuple
class PatchContents(_PatchContents):
    """The extracted data from a patch file

    :param readme: The contents of the README.txt file
    :type readme: String

    :param meta_ini: The contents of the meta.ini file
    :type meta_ini: String

    :param patched_files: A mapping of patched files to the updated content
    :type patched_files: Dictionary
    """
    pass


def extract_patch_contents(patch_path, log):
    """Convert the contents of a patch file into a usable data structure

    Yes, while this does read the whole patch into memory, it's highly unlikely
    that a patch for InsightIQ will be more than a dozen MB in size. By reading
    the patch into memory and generating a data structure for it, we exponentially
    increase out ability to unit test the patching logic.

    :Returns: TarContents (namedtuple)

    :param patch_path: **Required** The file system location of the patch file
    :type patch_path: String

    :param log: **Required** The logging object
    :type log: logging.Logger
    """
    log.debug("Validating patch at %s", patch_path)

    the_tar = tarfile.open(patch_path)
    tar_contents = the_tar.getmembers()
    meta_ini = None
    readme = None
    patched_files = {}
    for item in tar_contents:
        log.debug('Checking %s', item.path)
        if item.isdir():
            log.debug('Ignoring intermediate subdirectory')
            continue
        elif os.path.basename(item.path) == 'meta.ini':
            meta_ini = the_tar.extractfile(item.path).read()
            log.debug('Patch meta.ini contents:\n %s', meta_ini)
        elif os.path.basename(item.path) == 'README.txt':
            readme = the_tar.extractfile(item.path).read()
            log.debug('Patch README.txt found')
        else:
            patched_files[item.path] = the_tar.extractfile(item.path).read()
            log.debug('Added source file %s', item.path)
    the_tar.close()
    return PatchContents(readme, meta_ini, patched_files)


################################################
# Validators - Ensure and enforce requirements #
################################################

def check_file(cli_value):
    """Validate that the supplied value is actually a tar file that can be read

    This function is intended to be used with the argparse lib as an
    `argument type <https://docs.python.org/2.7/library/argparse.html#type>`_.

    :Raises: argparse.ArgumentTypeError

    :Returns: String

    :param cli_value: **Required** The value supplied by the end user
    :type cli_value: String
    """
    try:
        if not tarfile.is_tarfile(cli_value):
            msg = 'Supplied value is not a tar file: %s' % cli_value
            raise argparse.ArgumentTypeError(msg)
    except IOError as doh:
        # doesn't exist, or we can't read it
        msg = '%s: %s' % (doh.strerror, doh.filename)
        raise argparse.ArgumentTypeError(msg)
    else:
        return cli_value


def parse_cli(the_cli_args):
    """Defines the CLI for iiq_patch

    :Returns: argparse.Namespace

    :param the_cli_args: **Required** The commands from the CLI
    :type the_cli_args: List
    """
    parser = argparse.ArgumentParser(description='For installing/removing patches to InsightIQ',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    mutually_exclusive = parser.add_mutually_exclusive_group(required=True)
    mutually_exclusive.add_argument('-s', '--show', const='', nargs='?',
        help='Display patches that are installed. Supply value for info about that patch.')
    mutually_exclusive.add_argument('-i', '--install', type=check_file,
        help='The patch file you want to install')
    mutually_exclusive.add_argument('-u', '--uninstall',
        help='The name of the patch to remove')

    parser.add_argument('-d', '--debug', action='store_true',
        help='Increase the log level for iiq_patch')

    args = parser.parse_args(the_cli_args)
    return args


def patch_is_valid(contents, log):
    """Verify that the supplied patch file contents are OK

    :Returns: Boolean

    :param contents: **Required** The extracted patch file data
    :type contents: PatchContents

    :param log: **Required** The logging object
    :type log: logging.Logger
    """
    # Validate patch contained all expected data
    if not (contents.readme and contents.meta_ini and contents.patched_files):
        log.debug("patch is missing required data")
        return False

    # Validate the meta.ini contents
    config = ConfigObj(StringIO(contents.meta_ini))
    config_headers = set(config.keys())
    log.debug("meta.ini headers: %s", config_headers)
    if not config_headers == set(['info', 'version', 'files']):
        return False

    config_info_fields = set(config['info'].keys())
    log.debug("meta.ini info fields: %s", config_info_fields)
    if not config_info_fields == set(['name', 'bug']):
        return False

    config_version_fields = set(config['version'].keys())
    log.debug("meta.ini version fields: %s", config_version_fields)
    if not config_version_fields == set(['minimum', 'maximum']):
        return False

    if not len(config['files'].keys()) >= 1:
        log.debug('Supplied patch does not define what files to patch')
        return False
    for file_path in config['files'].keys():
        if file_path.startswith('/'):
            log.debug('Invalid file location for %s', file_path)
            return False
    return True


def source_is_patchable(files_to_patch, iiq_dir, log):
    """Compare the md5 hashes supplied in the patch with the md5 hashes of the
    currently installed source files.

    This function enables us to avoid installing a new patch that clobbers and
    existing patch; i.e. patch-b overwrites a file patched by patch-a.

    :Returns: Boolean

    :param files_to_patch: **Required** The mapping of file location to expected md5 hash of the source
    :type files_to_patch: Dictionary

    :param iiq_dir: **Required** The file system path where InsightIQ source is located.
    :type iiq_dir: String

    :param log: **Required** The logging object
    :type log: logging.Logger
    """
    is_patchable = True
    for the_file, the_hash in files_to_patch.items():
        abs_location = join_path(iiq_dir, the_file)
        log.debug('Checking IIQ source file %s', abs_location)
        log.debug("Expected hash %s for %s", the_hash, abs_location)
        try:
            md5_ok = md5_matches(abs_location, the_hash, log)
        except (IOError, OSError) as doh:
            log.error(doh)
            is_patchable = False
            break
        else:
            if not md5_ok:
                is_patchable = False
                break
            else:
                log.debug("Hashes match for %s", abs_location)
    return is_patchable


def version_ok(version_info, log):
    """Ensure the patch applies to the installed version of InsightIQ

    :Returns: Boolean

    :param version_info: **Required** Contains the minimum and maximum versions
                         of InsightIQ that the patch applies to.
    :type version_info: Dictionary

    :param log: **Required** The logging object
    :type log: logging.Logger
    """
    iiq_version = versions.get_iiq_version()
    min_version = versions.Version(name='patch-min', version=version_info['minimum'])
    max_version = versions.Version(name='patch-max', version=version_info['maximum'])
    log.info('Installed IIQ version: %s', iiq_version.version)
    log.info('Patch min version: %s', min_version.version)
    log.info('Patch max version: %s', max_version.version)
    return min_version <= iiq_version <= max_version


def md5_matches(file_location, md5_hash, log):
    """Determines if a supplied md5 hex matches the md5 for a given file

    :Returns: Boolean

    :param file_location: The file system path of the file to check
    :type file_location: String

    :param md5_hash: The expected md5 hash for the file, in hexadecimal format
    :type md5_hash: String

    :param log: **Required** The logging object
    :type log: logging.Logger
    """
    with open(file_location) as the_file:
        data = the_file.read()
    hasher = hashlib.md5()
    hasher.update(data)
    file_hash = hasher.hexdigest()
    log.debug('Found hash %s for %s', file_hash, file_location)
    return md5_hash == file_hash


def expected_backups(found_backups, expected_backups, log):
    """Verify the names and count of backed-up source files matches the original meta.ini

    :Returns: Boolean

    :param found_backups: **Required** The contents of the patch directory (i.e. listdir)
    :type found_backups: List

    :param expected_backups: **Required** The list of files that were patched (based of meta.ini)
    :type expected_backups: List

    :param log: **Required** The logging object
    :type log: logging.Logger
    """
    backup_files = set(found_backups)
    expected = set([convert_patch_name(x, to='backup') for x in expected_backups])
    log.debug("Found backups: %s", backup_files)
    log.debug("Expected backups: %s", expected_backups)
    if backup_files > expected:
        log.error("Found unexpected copies of original source files")
        return False
    elif backup_files < expected:
        log.error("Missing backups of original source files")
        return False
    elif backup_files != expected:
        log.error("Mismatching files between backups and expected backups")
        return False
    elif len(backup_files) != len(found_backups):
        log.error("Duplicate copies of backups found")
        return False
    else:
        return True


###################################################################
# Utility functions - Centralize conventions to avoid duplication #
###################################################################

def convert_patch_name(name, to='source'):
    """Coverts between the directory naming pattern, and the backup name.

    The patch installation creates backups of the source files. To avoid re-creating
    the directory structure of the source for the backup, we replace the forward
    slash ``/`` with a double-underbar ``__``. The double-underbar prevents us from
    accidentally converting ``my_module.py`` to ``my/module.py``

    Convert source file to backup name::

        >>> convert_patch_name('insightiq/controllers/security.py', to='backup')
        >>> 'insightiq__controllers__security.py'

    Convert backup to source name::

        >>> convert_patch_name('insightiq__controllers__security.py', to='source')
        >>> 'insightiq/controllers/security.py'

    :Returns: String

    :Raises: ValueError

    :param name: **Required** The name to convert
    :type name: String

    :param to: The target naming pattern
    :type to: Enum -> 'source', 'backup'
    """
    if to == 'source':
        return name.replace('___', '/')
    elif to == 'backup':
        return name.replace('/', '___')
    else:
        raise ValueError('param "to" must be either "source" or "backup", supplied: %s' % to)


def patch_backups_path(patch_dir):
    """Ceneralizes how other functions can reference the backups of original source files

    :Returns: String

    :param patch_dir: **Required** The file system path to the specific patch directory
    :type patch_dir: String
    """
    return '%s/originals' % patch_dir


def join_path(*args):
    """Join an arbitrary number of strings as a Linux/Unix file path

    Used in place of ``os.path.join`` so tests can run on Windows or Linux
    without generating false positives.

    Example::
        >>> join_path('some', 'directory')
        some/directory
        >>> join_path('/home', )

    :Returns: String

    :params args: **Required** The strings to join together
    :type args: String
    """
    return '/'.join(args)


#########################################################################
# Command handlers - carries out the procedure for the specific command #
#########################################################################

def restore_originals(iiq_dir, patch_dir, backup_files):
    """Replaces the patched source files with the backed up copies

    :Returns: None

    :param iiq_dir: The file system path to the InsightIQ source code
    :type iiq_dir: String

    :param patch_dir: The file system path to the patches directory
    :type patch_dir: String

    :param backup_files: The list of valid copies of the original source files
    :type backup_files: List
    """
    backups_dir = patch_backups_path(patch_dir)
    for backup in backup_files:
        source_name = convert_patch_name(backup, to='source')
        source_path = join_path(iiq_dir, source_name)
        backup_path = join_path(backups_dir, backup)
        source_pyc = source_path + 'c'
        shutil.copyfile(backup_path, source_path)
        if os.path.isfile(source_pyc):
            os.remove(source_pyc)


def install_patch(patch_content, patch_info, patch_name, log):
    """Does the actual installation of the InsightIQ patch

    :Returns: None

    :param patch_content: **Required** The extracted data from the patch file.
    :type patch_content: PatchContent (namedtuple)

    :param patch_info: **Required** The current state of patches installed.
    :type patch_info: PatchInfo (namedtuple)

    :param patch_name: **Required** The human reference to this patch.
    :type patch_name: String

    :param log: **Required** The logging object
    :type log: logging.Logger
    """
    patch_dir = join_path(patch_info.patches_dir, patch_name)
    patch_dir_backups = patch_backups_path(patch_dir)
    os.mkdir(patch_dir)
    os.mkdir(patch_dir_backups)
    # back up all source files 1st
    for file_location in patch_content.patched_files.keys():
        # no need to re-recreate directory structure under the patch dir for the backup
        backup_name = convert_patch_name(file_location, to='backup')
        backup_copy = join_path(patch_dir_backups, backup_name)
        original_copy = join_path(patch_info.iiq_dir, file_location)
        shutil.copyfile(original_copy, backup_copy)

    # Add README.txt to patch directory
    readme_location = join_path(patch_dir, 'README.txt')
    log.debug('Writing readme to %s', readme_location)
    with open(readme_location, 'w') as readme_file:
        readme_file.write(patch_content.readme)

    # Add meta.ini to patch directory
    meta_location = join_path(patch_dir, 'meta.ini')
    log.debug("Writing meta.ini to %s", meta_location)
    with open(meta_location, 'w') as meta_file:
        meta_file.write(patch_content.meta_ini)

    # Overwrite original source with patched version
    for patched_file, patched_data in patch_content.patched_files.items():
        location = join_path(patch_info.iiq_dir, patched_file)
        with open(location, 'w') as the_file:
            the_file.write(patched_data)
        # Delete any .pyc files
        if location.endswith('.py'):
            pyc = location + 'c'
            if os.path.isfile(pyc):
                os.remove(pyc)


def handle_show(specific_patch, log):
    """Implements the ``--show`` logic for iiq_patch.

    This function must be able to either display all installed patches, or
    details for a specific patch. The return value from the function must be
    the expected exit code for the iiq_patch script (i.e. non-zero for errors).

    :Returns: Integer

    :param specific_patch: **Required** The name of a specific patch to obtain
                           details for. Supply an zero length string to print
                           all installed patches to the console.
    :type specific_patch: String

    :param log: **Required** The logging object
    :type log: logging.Logger
    """
    exit_code = 0
    patch_info = versions.get_patch_info(specific_patch, log)
    log.debug(patch_info)
    if not patch_info.all_patches:
        log.info('No patches installed')

    elif specific_patch:
        log.debug('Looking up info for patch %s', specific_patch)
        if not patch_info.is_installed:
            log.error('Supplied patch does not exist: %s', patch_info.specific_patch)
            exit_code = 50
        elif not patch_info.readme:
            log.error('Unable to find patch details for %s', patch_info.specific_patch)
            exit_code = 51
        else:
            log.info('Details for patch %s\n', patch_info.specific_patch)
            print(patch_info.readme, '\n')
    else:
        # just wants a list of all patches
        log.debug("All patches: %s", patch_info.all_patches)
        msg = '\n\tPatches\n\t-------'
        msg += '\n\t%s\n' % '\n\t'.join(patch_info.all_patches)
        msg += '\n\tCount: %s\n' % len(patch_info.all_patches)
        print(msg)

    return exit_code


def handle_install(patch_path, log):
    """Entry point for installing a patch in InsightIQ.

    The returned integer is the intended exit code for the iiq_patch script.
    All explicitly defined exit codes (i.e. ones we don't bubble up from the system)
    must start at 100, and cannot exceed 199.

    :Returns: Integer

    :param patch_path: The file system location of the patch file
    :type patch_path: String

    :param log: The logging object
    :type log: logging.Logger
    """
    patch_info = versions.get_patch_info(specific_patch='', log=log)
    if not patch_info.iiq_dir:
        # IIQ is not installed, no way to install patch
        return 100

    # Prep system for patch installation
    try:
        os.mkdir(patch_info.patches_dir)
    except (OSError, IOError) as doh:
        if doh.errno == 13:
            log.error('Unable to setup patching directory')
            log.error("Please re-run with root privileges")
            return doh.errno
        elif doh.errno == 17:
            # dir already exists; likely already has a patch installed
            pass
        else:
            log.error(doh)
            if log.level == 10:
                log.exeception(doh)
            return doh.errno

    # Argprase should already have verified we can read the patch file
    # so no need to try/except for IO/OS errors
    patch_content = extract_patch_contents(patch_path, log)
    if not patch_is_valid(patch_content, log):
        log.error('Malformed patch file supplied')
        return 101
    meta_config = ConfigObj(StringIO(patch_content.meta_ini))
    if meta_config['info']['name'] in patch_info.all_patches:
        log.info('Patch %s is already installed', meta_config['info']['name'])
        return 0
    if not version_ok(meta_config['version'], log):
        log.error('Supplied patch does not apply to current version of InsightIQ')
        return 102
    if not source_is_patchable(meta_config['files'], patch_info.iiq_dir, log):
        log.error("Unable to install patch. Please contact Isilon support.")
        return 103
    try:
        install_patch(patch_content=patch_content,
                      patch_info=patch_info,
                      patch_name=meta_config['info']['name'],
                      log=log)
    except (IOError, OSError) as doh:
        specific_patch_dir = join_path(patch_info.patches_dir, meta_config['info']['name'])
        # ignore_errors encase the patch specific dir was never made
        shutil.rmtree(specific_patch_dir, ignore_errors=True)
        log.error(doh)
        log.info('Unable to install patch')
        if log.level == 10:
            # i.e. debug logging
            log.exception(doh)
        return 104
    else:
        log.info('Successfully installed patch')
    return 0


def handle_uninstall(patch_name, log):
    """Carries out the task of uninstalling a patch

    The returned integer is the exit code for the script. All explicitly defined
    exit codes (i.e. ones we don't bubble up from the system) must start at
    200, and cannot exceed 255.

    :Returns: Integer

    :param patch_name: **Required** The name of the patch to uninstall
    :type patch_name: String

    :param log: **Required** The logging object.
    :type log: logging.Logger
    """
    patch_info = versions.get_patch_info(specific_patch=patch_name, log=log)
    patch_dir = join_path(patch_info.patches_dir, patch_name)
    if not patch_info.is_installed:
        log.info('Patch %s is not installed', patch_name)
        return 200

    meta_location = join_path(patch_dir, 'meta.ini')
    try:
        meta_config = ConfigObj(meta_location)
    except (IOError, OSError) as doh:
        log.error(doh)
        log.info('Unable to uninstall patch %s', patch_name)
        return 201

    backups_dir = patch_backups_path(patch_dir)
    backup_files = os.listdir(backups_dir)
    if not expected_backups(backup_files, meta_config['files'].keys(), log):
        log.error("Unable to remove patch")
        return 202

    for original_path, original_hash in meta_config['files'].items():
        backup_name = convert_patch_name(original_path, to='backup')
        backup_file = join_path(backups_dir, backup_name)
        if not md5_matches(backup_file, original_hash, log):
            return 203

    try:
        restore_originals(patch_info.iiq_dir, patch_dir, backup_files)
    except (IOError, OSError) as doh:
        log.error("Unable to complete restore of original source files")
        log.error(doh)
        return doh.errno

    try:
        shutil.rmtree(patch_dir)
    except (IOError, OSError) as doh:
        log.error("Unable to dereference patch: %s", patch_name)
        log.error(doh)
        log.error('Please manually delete: %s', patch_dir)
        return doh.errno
    else:
        log.info("Successfully uninstalled patch")
    return 0


def main(the_cli_args):
    """Entry point function for iiq_patch"""
    args = parse_cli(the_cli_args)
    if args.debug:
        log_level = 10
    else:
        log_level = 20
    log = get_logger(log_path='/dev/null', stream_lvl=log_level, file_lvl=log_level)

    if args.install:
        exit_code = handle_install(args.install, log)
    elif args.uninstall:
        exit_code = handle_uninstall(args.uninstall, log)
    else:
        exit_code = handle_show(args.show, log)
    log.debug('Exit code: %s', exit_code)
    if (not exit_code) and (args.install or args.uninstall):
        # successfully installed or uninstalled the patch
        restart_syntax = 'service insightiq restart'
        if getpass.getuser() == 'root':
            log.info('Restarting InsightIQ')
            result = shell.run_cmd(restart_syntax)
            log.info(result.stdout)
            log.debug('Sterr from retart syntax: %s', result.stderr)
            log.debug('Restart syntax: %s', result.command)
            log.debug('Restart exitcode: %s', result.exit_code)
        else:
            log.info('Non-root user detected for patch install. Unable to restart InsightIQ.')
            log.info('**Patch wont take effect unless you restart InsightIQ**\n')
            log.info('Please run the following command to restart InsightIQ:')
            log.info('sudo %s', restart_syntax)
    return exit_code



# -*- coding: UTF-8 -*-
"""
This module make obtaining and comparing version strings easy!
"""
import glob
import os
from collections import namedtuple
from pkg_resources import get_distribution, DistributionNotFound


class Version(object):
    """Implements comparison operators for common version strings

    Only versions strings upwards of 4 digits and consisting only of numbers is
    supported. Version strings breakdown into ``major``, ``minor``, ``patch``,
    and ``build``.

    Example::

        >>> version_as_string = '1.2.3'
        >>> version = Version(name='myPackage', version=version_as_string)
        >>> version.major
        1
        >>> version.patch
        3
        >>> type(version.minor)
        <type 'int'>
        >>> type(version.build)
        <type 'NoneType'>

    Comparing Versions::

        >>> v1 = Version(name='myPackage', version='1.2.3')
        >>> v2 = Version(name='newPackage', version='1.4.0')
        >>> v1 < v2
        True
        >>> v2 >= v1
        True
        >>> v2 == v1
        False

    The Version object also support comparison of string versions::

        >>> v1 = Version(name='foo', version='4.5')
        >>> v1 > '5.0'
        False

    It's worth noting, more specific versions (that would otherwise be equal),
    are considered greater::

        >>> v1 = Version(name='bar', version='1.2.0')
        >>> v1 > '1.2'
        True

    This is because a version without a value is None, and zero is greater than None in Python::

       >>> 0 > None
       True
    """

    # for semantic versioning http://semver.org
    _semver = ('major', 'minor', 'patch', 'build')

    def __init__(self, version, name):
        self._name = name
        if not isinstance(version, str):
            msg = 'Version object can only be created from string, supplied %s, %s' % (version, type(version))
            raise TypeError(msg)
        else:
            version_breakdown = version.split('.')
            if len(version_breakdown) < 2:
                # Pretty much every version number is dot-delimited so we should
                # at least two items, must not be a version
                raise ValueError("Unexpected value for version, %s" % version)
            else:
                for idx, semantic_version in enumerate(self._semver):
                    try:
                        number = int(version_breakdown[idx])
                        setattr(self, '_%s' % semantic_version, number)
                    except ValueError:
                        msg = 'Version only support integer values, failed to cast %s for %s' % (semantic_version, version)
                        raise ValueError(msg)
                    except IndexError:
                        if idx == 3 or idx == 2:
                            # some versions only consist of major.monior, like google chrome
                            setattr(self, '_%s' % semantic_version, None)
                        else:
                            # re-raise the error for debugging
                            raise

        self._version = version

    @property
    def name(self):
        return self._name

    @property
    def semver(self):
        return self._semver

    @property
    def version(self):
        return self._version

    @property
    def major(self):
        return self._major

    @property
    def minor(self):
        return self._minor

    @property
    def patch(self):
        return self._patch

    @property
    def build(self):
        return self._build

    def __len__(self):
        return len(self._version.split('.'))

    def __repr__(self):
        """How we represent the object"""
        return 'Version(name=%s, version=%s)' % (self.name, self.version)

    def _get_other(self, other):
        """This method is to enable comparison with versions as strings

        :Returns: Version

        :param other: **Required** The object being compared to this one
        :type other: String or Version
        """
        if isinstance(other, str):
            try:
                other_version = Version(other, name=None)
            except ValueError:
                msg = "Unable to compare %s and %s" % self, other
                raise TypeError(msg)
            else:
                return other_version
        elif isinstance(other, Version):
            return other
        else:
            msg = 'Unable to compare types %s and %s' % (type(self), type(other))
            raise TypeError(msg)

    def __hash__(self):
        """Because __hash__ goes away if you define __eq__"""
        return hash(self._version)

    def __eq__(self, other):
        """Defines the behavior for testing for equivalence, '=='

        :Returns: Boolean

        :param other: The other object to compare this one against
        :type other: Version
        """
        other_version = self._get_other(other)
        for semver in self._semver:
            if getattr(self, semver) != getattr(other_version, semver):
                return False
        else:
            # Yup, for/else is a thing in Python: http://book.pythontips.com/en/latest/for_-_else.html
            return True

    def __ne__(self, other):
        """Defines the behavior for testing for non-equivalence, '!='

        :Returns: Boolean

        :param other: The other object to compare this one against
        :type other: Version
        """
        other_version = self._get_other(other)
        for semver in self._semver:
            if getattr(self, semver) != getattr(other_version, semver):
                return True
        else:
            return False

    def __lt__(self, other):
        """Define the behavior for testing for less than '<'

        :Returns: Boolean

        :param other: The other object to compare this one against
        :type other: Version
        """
        other_version = self._get_other(other)
        for semver in self._semver:
            if getattr(self, semver) > getattr(other_version, semver):
                return False
            elif getattr(self, semver) < getattr(other_version, semver):
                return True

        if len(self) == len(other_version):
            return False

    def __gt__(self, other):
        """Define the behavior for testing for greater than or equal, '>'

        :Returns: Boolean

        :param other: The other object to compare this one against
        :type other: Version
        """
        result = True
        other_version = self._get_other(other)
        for semver in self._semver:
            if getattr(self, semver) > getattr(other_version, semver):
                return True
            elif getattr(self, semver) < getattr(other_version, semver):
                return False

        if len(self) == len(other_version):
            return False

    def __le__(self, other):
        """Define the behavior for testing for less than or equal '<='

        :Returns: Boolean

        :param other: The other object to compare this one against
        :type other: Version
        """
        other_version = self._get_other(other)
        for semver in self._semver:
            if getattr(self, semver) < getattr(other_version, semver):
                return True
            elif getattr(self, semver) > getattr(other_version, semver):
                return False

        if len(self) == len(other_version):
            return True

    def __ge__(self, other):
        """Define the behavior for testing for greater than, '>='

        :Returns: Boolean

        :param other: The other object to compare this one against
        :type other: Version
        """
        other_version = self._get_other(other)
        for semver in self._semver:
            if getattr(self, semver) > getattr(other_version, semver):
                return True
            elif getattr(self, semver) < getattr(other_version, semver):
                return False

        if len(self) == len(other_version):
            return True


def get_iiq_version():
    """Obtain the version of InsightIQ installed

    :Returns: iiqtools.utils.versions.Version
    """
    try:
        iiqtools_version = get_distribution('isilon_insightiq').version
    except DistributionNotFound:
        return None
    else:
        return Version(name='insightiq', version=iiqtools_version)


def get_iiqtools_version():
    """Obtain the version of iiqtools installed

    :Returns: iiqtools.utils.versions.Version
    """
    try:
        iiqtools_version = get_distribution('iiqtools').version
    except DistributionNotFound:
        return None
    else:
        return Version(name='iiqtools', version=iiqtools_version)


_PatchInfo = namedtuple("PatchInfo", "iiq_dir patches_dir specific_patch is_installed readme all_patches")
# So we can have a nice docstring for the namedtuple
class PatchInfo(_PatchInfo):
    """Describes the state of patches for InsightIQ

    :param iiq_dir: The file system path where InsightIQ source is located.
    :type iiq_dir: String

    :param patches_dir: The file system path where patches for InsightIQ are stored.
    :type patches_dir: String

    :param specific_patch: Only populated when a patch is being installed/removed/read.
    :type specific_patch: String

    :param is_installed: If ``specific_patch`` is installed or not.
    :Type is_installed: Boolean

    :param readme: The README.txt for ``specific_patch`` if applicable.
    :type readme: String

    :param all_patches: All currently installed patches.
    :type all_patches: Tuple
    """
    pass


def get_patch_info(specific_patch, log):
    """Obtain the current state of patches for InsightIQ

    :Returns: PatchInfo (namedtuple)

    :param specific_patch: **Required** The name of a patch that's being installed/removed/read.
    :type specific_patch: String

    :param log: **Required** The logging object. This param is really here to make unit testing
                easier -> https://en.wikipedia.org/wiki/Dependency_injection
    :type log: logging.Logger
    """
    try:
        # IIQ 4.1.0 and newer runs on Python 2.7, older versions use  Python 2.6
        iiq_dir = glob.glob('/usr/share/isilon/lib/python2.*/site-packages')[0]
    except IndexError:
        log.debug('Unable to find InsightIQ install dir. Is it installed?')
        iiq_dir = ''
    patches_dir = iiq_dir + '/' + 'insightiq/patches'
    try:
        all_patches = tuple(os.listdir(patches_dir))
    except (OSError, IOError) as doh:
        log.debug('Unable to list %s', patches_dir)
        all_patches = []

    is_installed = specific_patch in all_patches
    try:
        with open(patches_dir + '/' + specific_patch + '/' + 'README.txt') as the_file:
            readme = the_file.read()
    except (OSError, IOError) as doh:
        if specific_patch:
            log.debug('%s : %s', doh.strerror, doh.filename)
        readme = ''

    return PatchInfo(iiq_dir, patches_dir, specific_patch, is_installed, readme, all_patches)



# -*- coding: UTF-8 -*-
"""
Unit tests for the iiqtools.utils.cli_parsers module
"""
import os
import json
import unittest

from iiqtools.utils import cli_parsers


def find_examples_dir():
    """Helper function so we can find the example output files regardless from
    where this script runs
    """
    current_location = os.path.abspath(__file__)
    parent_dir = os.path.dirname(current_location)
    return os.path.join(parent_dir, 'cli_parser_examples')

EXAMPLES_DIR = find_examples_dir()


# Individual test cases makes it easy to test just one parser
class TestIfconfigToDict(unittest.TestCase):
    """A suite of test cases for the iiqtools.utils.cli_parsers.ifconfig_to_dict function"""

    def test_ifconfig_to_dict_1(self):
        """Example 1 of `ifconfig` cli output parses to expected structure"""
        with open(os.path.join(EXAMPLES_DIR, 'ifconfig1.json')) as the_file:
            data = json.load(the_file)
        example_output = data['stdout']
        parsed = cli_parsers.ifconfig_to_dict(example_output)

        result = cmp(parsed['interfaces'], data['interfaces'])
        expected = 0 # cmp return zero when values are identical
        self.assertEqual(result, expected)


class TestDfToDict(unittest.TestCase):
    """A suite of test cases for the iiqtools.utils.cli_parsers.df_to_dict function"""

    def test_df_to_dict_1(self):
        """Example 1 of `df` cli output parses to expected structure"""
        with open(os.path.join(EXAMPLES_DIR, 'df1.json')) as the_file:
            data = json.load(the_file)
        example_output = data['stdout']
        parsed = cli_parsers.df_to_dict(example_output)

        result = cmp(parsed['filesystems'], data['filesystems'])
        expected = 0 # cmp return zero when values are identical
        self.assertEqual(result, expected)


class TestMemoryToDict(unittest.TestCase):
    """A suite of test cases for the iiqtools.utils.cli_parsers.memory_to_dict function"""

    def test_memory_to_dict_1(self):
        """Example 1 of `free` cli output parses to expected structure"""
        with open(os.path.join(EXAMPLES_DIR, 'memory1.json')) as the_file:
            data = json.load(the_file)
        example_output = data['stdout']
        parsed = cli_parsers.memory_to_dict(example_output)

        result = cmp(parsed['memory'], data['memory'])
        expected = 0 # cmp return zero when values are identical
        self.assertEqual(result, expected)


if __name__ == '__main__':
    unittest.main()



# -*- coding: UTF-8 -*-
"""
This module contains all the business logic for collecting logs and configuration
information about InsightIQ for remote troubleshooting.
"""
import os
import sys
import time
import json
import glob
import tarfile
import argparse
import traceback
import subprocess
from getpass import getuser
from StringIO import StringIO

try:
    from iiq_data_export.api_connection import iiq_api
    from insightiq.lib.api_connection.api import APIConnectionError
except ImportError:
    # InsightIQ is closed source, so it cannot be included as a dev/test dependency.
    # We're stuck with assuming a dev/test environment, and mocking it away.
    # This is the only way we can do unit testing, and yes, it's a bit hacky.
    from mock import MagicMock
    iiq_api = MagicMock()
    APIConnectionError = RuntimeError

from iiqtools.utils import versions
from iiqtools.utils import cli_parsers
from iiqtools.utils.logger import get_logger
from iiqtools.utils.generic import check_path
from iiqtools.utils.shell import run_cmd
from iiqtools.exceptions import CliError


def get_tarfile(output_dir, case_number, the_time=None):
    """Centralizes logic for making tgz file for InsightIQ logs

    :Returns: tarfile.TarFile

    :param output_dir: **Required** The directory to save the tar file in
    :type output_dir: String

    :param case_number: **Required** The SR that the logs are for; used in file name.
    :type case_number: String or Integer

    :param the_time: An optional EPOC timestamp to use when naming the file.
                     If not supplied, this function calls time.time().
    :type the_time: EPOC time stamp
    """
    base_dir = os.path.abspath(output_dir)
    if the_time is None:
        the_time = int(time.time())
    file_name = 'IIQLogs-sr%s-%s.tgz' % (case_number, the_time)
    full_path = os.path.join(base_dir, file_name)
    return tarfile.open(full_path, 'w:gz') # 2nd param makes it a tgz file


def add_from_memory(the_tarfile, data_name, data):
    """Simplify adding in-memory information to the tar file

    :Returns: None

    :param the_tarfile: The open tarfile object
    :type the_tarfile: tarfile.open

    :param data_name: The reference to the data; i.e. it's name when you uncompress the file
    :type data_name: String

    :param data: The contents of the in-memory information
    :type data: String
    """
    info = tarfile.TarInfo(data_name)
    info.size = len(data)
    info.mtime = time.time()
    info.mode = int('444', 8) # everyone can read in oct, not decimal
    the_tarfile.addfile(info, StringIO(data))


def parse_cli(cli_args):
    """Handles parsing the CLI, and gives us --help for (basically) free

    :Returns: argparse.Namespace

    :param cli_args: The arguments passed to the script
    :type cli_args: List
    """
    parser = argparse.ArgumentParser(description='Generate a .tar file for debugging InsightIQ',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('--output-dir', default='/home/administrator', type=check_path,
        help='The directory to write the .tgz file to')
    parser.add_argument('--case-number', required=True,
        help='The Service Request number. Used in naming tar file.')

    args = parser.parse_args(cli_args)
    return args


# IIQ API info
def call_iiq_api(uri):
    """Make an internal API call to the InsightIQ API

    :Returns: JSON String

    :param uri: The API end point to call
    :type uri: string
    """
    template = {'category' : 'api',
                'endpoint' : uri,
                'error' : '',
                'response' : None,
                'traceback' : ''}
    try:
        resp = iiq_api.make_request(uri)
    except APIConnectionError:
        template['error'] = "unable to query InsightIQ API"
    except Exception as doh:
        template['error'] = "%s" % doh
        template['traceback'] = traceback.format_exc()
    else:
        data = json.loads(resp.read())
        template['response'] = data

    # The indent makes the JSON more human friendly, which GS appreciates
    return json.dumps(template, indent=4, sort_keys=True)


def datastore_info():
    """Obtain data about the datastore for InsightIQ

    :Returns: JSON String
    """
    return call_iiq_api('/api/datastore/usage?current_dir=true')


def clusters_info():
    """Obtain a pile of data about all monitored clusters in InsightIQ

    :Returns: JSON String
    """
    return call_iiq_api('/api/clusters')


def ldap_info():
    """Obtain the config for LDAP in InsightIQ

    :Returns: JSON String
    """
    return call_iiq_api('/api/ldap/configs')


def reports_info():
    """Obtain info about any scheduled reports in InsightIQ

    :Returns: JSON String
    """
    return call_iiq_api('/api/reports')


# CLI command info
def cli_cmd_info(command, parser):
    """Standardizes the JSON format for any data collected via a CLI command

    :Returns: JSON String

    :param command: The CLI command to execute
    :type command: String
    """
    try:
        result = run_cmd(command)
    except CliError as result:
        pass
    template = {'exitcode' : result.exit_code,
                'stderr' : result.stderr,
                'command' : result.command,
                'stdout' : result.stdout,
                'category' : 'cli',
                'traceback' : ''}
    if not (result.exit_code or result.stderr):
        # I know, shitty heuristic to determine errors...
        try:
            parsed_format = parser(result.stdout)
        except Exception:
            template['traceback'] = traceback.format_exc()
        else:
            template.update(parsed_format)

    return json.dumps(template, indent=4, sort_keys=True)


def mount_info():
    """Obtain data about mounted file systems on the host OS

    :Returns: JSON String
    """
    return cli_cmd_info('df -P', cli_parsers.df_to_dict)


def memory_info():
    """Obtain data about RAM on the host OS

    :Returns: JSON String
    """
    return cli_cmd_info('free -m', cli_parsers.memory_to_dict)


def ifconfig_info():
    """Obtain data about the network interfaces on the host OS

    :Returns: JSON String
    """
    return cli_cmd_info('ifconfig', cli_parsers.ifconfig_to_dict)


def iiq_version_info():
    """Obtain info about the version of InsightIQ installed on the host OS

    :Returns: JSON String
    """
    iiq_version = versions.get_iiq_version()
    iiqtools_version = versions.get_iiqtools_version()
    data = {'insightiq' : iiq_version.version,  'iiqtools' : iiqtools_version.version}
    return json.dumps(data, indent=4, sort_keys=True)


def main(the_cli_args):
    """Entry point for running script"""
    args = parse_cli(the_cli_args)
    user = getuser()
    if user != 'root':
        warning = '\n****WARNING****\nNot all data is going to be collected because\n'
        warning += 'some log files can only be read by root.\n'
        warning += 'To collect all diagnostic data, you must run as root.\n'
        warning += 'You are currently running as %s\n' % user
        print(warning)
        response = raw_input('Do you wish to continue? (yes/no): ')
        if not response.lower().startswith('y'):
            return 1

    log = get_logger(log_path='/dev/null', stream_lvl=10, file_lvl=10)
    try:
        tar_file = get_tarfile(args.output_dir, args.case_number)
    except (IOError, OSError) as doh:
        log.error(doh)
        return doh.errno

    log.info('Collecting config information')
    add_from_memory(tar_file, 'iiq_version.json', iiq_version_info())
    add_from_memory(tar_file, 'ifconfig.json', ifconfig_info())
    add_from_memory(tar_file, 'memory.json', memory_info())
    add_from_memory(tar_file, 'df.json', mount_info())

    add_from_memory(tar_file, 'ldap.json', ldap_info())
    add_from_memory(tar_file, 'clusters.json', clusters_info())
    add_from_memory(tar_file, 'datastore.json', datastore_info())

    # add release info files
    for releaseinfo in glob.glob('/etc/*release'):
        tar_file.add(releaseinfo)

    # add config files
    for a_file in glob.glob('/etc/isilon/*'):
        tar_file.add(a_file)
    for rabbitconfig in glob.glob('/etc/rabbitmq/*.config'):
        tar_file.add(rabbitconfig)

    log.info('Collecting log files')
    # add iiq log files
    for iiq_log in glob.glob('/var/log/insightiq*'):
        tar_file.add(iiq_log)
    for iiq_log in glob.glob('/var/log/iiq*'):
        tar_file.add(iiq_log)
    for iiq_log in glob.glob('/var/log/insightiq_clusters/*/*'):
        tar_file.add(iiq_log)

    # add postgres logs
    try:
        tar_file.add('/var/cache/insightiq/pgsql/pgstartup.log')
    except (IOError, OSError):
        # It's a permissions error; already prompted user that not all files
        # can be collected if not ran as root
        pass
    for pglog in glob.glob('/var/log/pg_log/*'):
        tar_file.add(pglog)

    # add rabbitmq logs
    for rabbitlog in glob.glob('/var/log/rabbitmq/*'):
        tar_file.add(rabbitlog)

    tar_file.close()
    log.info('Log gather complete')
    log.info('Created log file %s', tar_file.name)
    return 0



# -*- coding: UTF-8 -*-
"""This module reduces boilerplate when interacting with the command shell, i.e. BASH.

Example A::

    >>> result = shell.run_cmd('ls')
    >>> print result.stdout
    README.txt
    someOtherFile.txt
    >>> print result.stderr

    >>> result.exit_code
    0

Example B::

    >>> # run_cmd does not support the Unix Pipeline
    >>> result = shell.run_cmd('cat foo.txt | grep "not supported"')
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
        raise CliError(cli_syntax, stdout, stderr, exit_code)
    iiqtools.exceptions.CliError: Command Failure: cat foo.txt | grep "not supported"
"""
import subprocess
from collections import namedtuple

from iiqtools.exceptions import CliError

_CliResult = namedtuple('CliResult', 'command stdout stderr exit_code')
# Enables us to create a handy docstring for Sphinx
class CliResult(_CliResult):
    """The outcome from running a CLI command

    :Type: collections.namedtuple

    :param command: The CLI command that was ran
    :type command: String

    :param stdout: The output from the standard out stream
    :type stdout: String

    :param stderr: The output from the standard error stream
    :type stderr: String

    :param exit_code: The exit/return code from the command
    :type exit_codee: Integer
    """
    pass

def run_cmd(cli_syntax):
    """Execute a simple CLI command.

    This function blocks until the CLI command returns and does not support the
    Unix pipeline.

    :Returns: CliResult - namedtuple

    :Raises: CliError (when exit code is not zero)

    :param cli_syntax: The CLI command to run.
    :type cli_syntax: String
    """
    CliResult = namedtuple('CliResult', 'command stdout stderr exit_code')
    try:
        proc = subprocess.Popen(cli_syntax.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except OSError as doh:
        stdout = ''
        stderr = '%s' % doh
        exit_code = 1
    else:
        stdout, stderr = proc.communicate()
        exit_code = proc.returncode
    if exit_code:
        raise CliError(cli_syntax, stdout, stderr, exit_code)
    else:
        return CliResult(cli_syntax, stdout, stderr, exit_code)



# -*- UTF-8 -*-
"""
Unit tests for the iiqtools.utils.shell module
"""
import unittest

from iiqtools.utils import shell
from iiqtools.exceptions import CliError


class TestShell(unittest.TestCase):
    """A suite of test cases for the iiqtools.utils.shell module"""

    def test_happy_path(self):
        """Running a valid command returns an instance of shell.CliResult"""
        result = shell.run_cmd('ls')
        self.assertTrue(isinstance(result, tuple))

    def test_cli_error(self):
        """Running a command that has a non-zero exit code raises CliError"""
        self.assertRaises(CliError, shell.run_cmd, 'not a command')

    def test_cli_result_attributes(self):
        """Verify that the CliResult object attributes remain stable"""
        # NEVER REMOVE OR MODIFY A VALUE IN THIS LIST
        expected = ['command', 'exit_code', 'stderr', 'stdout']
        result = shell.run_cmd('ls')
        for attr in expected:
            self.assertTrue(hasattr(result, attr))


if __name__ == '__main__':
    unittest.main()



# -*- coding: UTF-8 -*-
import logging
import os.path


def get_logger(log_path=None, stream_lvl=0, file_lvl=logging.INFO):
    """Factory for making logging objects

    The verbosity of the logs are configurable as defined by the official Python
    documentation: https://docs.python.org/2/library/logging.html#logging-levels

    :Returns: logging.Logger

    :Raises: AssertionError on bad parameter input

    :param log_path: **Required** The absolute file path to write logs to
    :type log_path: String

    :param stream_lvl: Set to print log messages to the terminal.
    :type stream_lvl: Integer, default 0

    :param file_lvl: How verbose the log file messages are. This value cannot be zero.
    :type stream_lvl: Integer, default 20
    """
    # pylint: disable=line-too-long
    assert stream_lvl in (logging.NOTSET, logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL)
    # pylint: disable=line-too-long
    assert file_lvl in (logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL)
    assert isinstance(log_path, str) or isinstance(log_path, unicode)
    assert os.path.isabs(log_path)
    assert os.path.isdir(os.path.dirname(log_path))

    # Determine the most verbose log level so both file and stream messages
    # can be accepted by the logging object.
    # Cannot simply take min() of the params because a value of zero means "don't log"
    if stream_lvl:
        base_lvl = min(stream_lvl, file_lvl)
    else:
        base_lvl = file_lvl

    log = logging.getLogger(name=log_path)
    if log.handlers:
        # Calling for the same log multiple times would set multiple handlers
        # If you have 2 duplicate file handers, you write twice to the log file
        return log
    log.setLevel(base_lvl)
    formatter  = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')

    file_handler = logging.FileHandler(log_path)
    file_handler.setFormatter(formatter)
    file_handler.setLevel(file_lvl)

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(stream_lvl)

    log.addHandler(file_handler)
    log.addHandler(stream_handler)
    return log



# -*- coding: UTF-8 -*-
"""
Unit tests for the iiqtools.iiq_gather_info logic
"""
import os
import json
import gzip
import glob
import unittest
import argparse
import __builtin__

from mock import patch, MagicMock

from iiqtools import iiq_gather_info
from iiqtools.utils.versions import Version


class TestGetTarfile(unittest.TestCase):
    """A suite of test cases for the iiq_gather_info.get_tarfile function"""

    def setUp(self):
        """Runs before every test case"""
        self.output_dir = '/tmp'
        self.case_number = 1
        self.the_time = 1234

    def tearDown(self):
        """Runs after every test case"""
        test_files = glob.glob('/tmp/IIQLogs*')
        for the_file in test_files:
            os.remove(the_file)

    def test_get_tarfile_name(self):
        """The `get_tarfile` function returns the expected file"""
        the_tarfile = iiq_gather_info.get_tarfile(self.output_dir,
                                                  self.case_number,
                                                  the_time=self.the_time)
        tarfile_name = the_tarfile.name
        expected_name = '%s/IIQLogs-sr%s-%s.tgz' % (self.output_dir,
                                                        self.case_number,
                                                        self.the_time)

        self.assertEqual(tarfile_name, expected_name)

    def test_get_tarfile_is_compressed(self):
        """The `get_tarfile` function returns a gzipped file"""
        the_tarfile = iiq_gather_info.get_tarfile(self.output_dir,
                                                  self.case_number)

        self.assertTrue(isinstance(the_tarfile.fileobj, gzip.GzipFile))



class TestAddFromMemory(unittest.TestCase):
    """A suite of tests for the iiq_gather_info.add_from_memory function"""

    def setUp(self):
        """Runs before every test case"""
        self.fake_tarfile = MagicMock()

    def test_add_from_memory_addfile(self):
        """The add_from_memory function adds the data to the tar file"""

        iiq_gather_info.add_from_memory(the_tarfile=self.fake_tarfile,
                                        data_name='foo.json',
                                        data='{"some" : "json string"}')
        call_count = self.fake_tarfile.addfile.call_count
        expected = 1

        self.assertEqual(call_count, expected)

    @patch.object(iiq_gather_info.tarfile, 'TarInfo')
    def test_add_from_memory_mode(self, fake_tarinfo):
        """The add_from_memory fuction sets the correct POSIX permissions"""
        fake_info = MagicMock()
        fake_tarinfo.return_value = fake_info
        iiq_gather_info.add_from_memory(the_tarfile=self.fake_tarfile,
                                        data_name='foo.json',
                                        data='{"some" : "json string"}')
        posix_oct = fake_info.mode
        expected = 292 # that's 444 POSIX in decimal

        self.assertEqual(posix_oct, expected)


class TestParseCli(unittest.TestCase):
    """A suite of test cases for the parse_cli function"""

    def test_returns_namespace(self):
        """Supplying all required args to iiq_gather_info.parse_cli returns a namespace object"""
        sent_args = ['--output-dir', '/tmp', '--case-number', '2']
        out_args = iiq_gather_info.parse_cli(sent_args)

        self.assertTrue(isinstance(out_args, argparse.Namespace))

    # Mocking away stderr to avoid spam in output while running tests
    @patch.object(iiq_gather_info.argparse._sys, 'stderr')
    def test_no_args(self, fake_stderr):
        """Passing zero args to iiq_gather_info.parse_cli raises SystemExit"""
        sent_args = []
        self.assertRaises(SystemExit, iiq_gather_info.parse_cli, sent_args)

    @patch.object(iiq_gather_info.argparse._sys, 'stderr')
    def test_only_output_dir(self, fake_stderr):
        """Supplying only --output-dir raises SystemExit"""
        sent_args = ['--output-dir', '/tmp',]
        self.assertRaises(SystemExit, iiq_gather_info.parse_cli, sent_args)

    @patch.object(iiq_gather_info.argparse._sys, 'stderr')
    def test_only_case_number(self, fake_stderr):
        """Supplying only --case-number raises SystemExit"""
        sent_args = ['--case-number', '1',]
        self.assertRaises(SystemExit, iiq_gather_info.parse_cli, sent_args)


class TestCallIiqApi(unittest.TestCase):
    """A suite of test cases for the call_iiq_api function"""

    def setUp(self):
        """Runs before every test case"""
        # InsightIQ is closed source, so in the lib we replace the iiq_api
        # function (which we cannot import) with a mock object
        iiq_gather_info.iiq_api = MagicMock()
        self.fake_response = MagicMock()
        self.fake_response.read.return_value = '{"foo":"bar"}'
        iiq_gather_info.iiq_api.make_request.return_value = self.fake_response

    def test_response_is_json(self):
        """The response is seralized, valid JSON"""
        response = iiq_gather_info.call_iiq_api(uri='/foo')
        deseralized = json.loads(response)

        self.assertTrue(isinstance(response, str))
        self.assertTrue(isinstance(deseralized, dict))
        self.assertEqual(deseralized['response'], {'foo' : 'bar'})

    def test_api_error(self):
        """The response object notes if the IIQ API failed"""
        iiq_gather_info.iiq_api.make_request.side_effect = [RuntimeError('testing')]

        response = iiq_gather_info.call_iiq_api(uri='/foo')
        data = json.loads(response)
        err_msg = 'unable to query InsightIQ API'

        self.assertEqual(err_msg, data['error'])

    def test_notes_tracebacks(self):
        """Unexpected exceptions have the assocated traceback logged in the response"""
        # So we can A) fix it, and B) write a unit test for it! :D
        iiq_gather_info.iiq_api.make_request.side_effect = [Exception('testing')]

        response = iiq_gather_info.call_iiq_api(uri='/foo')
        data = json.loads(response)
        err_msg = 'testing'

        self.assertEqual(err_msg, data['error'])
        self.assertTrue(data['traceback'].startswith('Traceback'))


class TestIiqApiHelpers(unittest.TestCase):
    """A suite of test cases for the helper functions that pull
       data from the InsightIQ API
    """

    def setUp(self):
        """Runs before every test case"""
        # InsightIQ is closed source, so in the lib we replace the iiq_api
        # function (which we cannot import) with a mock object
        iiq_gather_info.iiq_api = MagicMock()
        self.fake_response = MagicMock()
        self.fake_response.read.return_value = '{"foo":"bar"}'
        iiq_gather_info.iiq_api.make_request.return_value = self.fake_response

    def test_datastore_info(self):
        """Function `datastore_info` calls the correct API endpoint"""
        data = iiq_gather_info.datastore_info()
        the_json = json.loads(data)
        expected_endpoint = '/api/datastore/usage?current_dir=true'

        self.assertEqual(the_json['endpoint'], expected_endpoint)

    def test_clusters_info(self):
        """Function `clusters_info` calls the correct API endpoint"""
        data = iiq_gather_info.clusters_info()
        the_json = json.loads(data)
        expected_endpoint = '/api/clusters'

        self.assertEqual(the_json['endpoint'], expected_endpoint)

    def test_ldap_info(self):
        """Function `ldap_info` calls the correct API endpoint"""
        data = iiq_gather_info.ldap_info()
        the_json = json.loads(data)
        expected_endpoint = '/api/ldap/configs'

        self.assertEqual(the_json['endpoint'], expected_endpoint)

    def test_reports_info(self):
        """Function `reports_info` calls the correct API endpoint"""
        data = iiq_gather_info.reports_info()
        the_json = json.loads(data)
        expected_endpoint = '/api/reports'

        self.assertEqual(the_json['endpoint'], expected_endpoint)


class TestCliCmdInfo(unittest.TestCase):
    """A suite of tests for the cli_cmd_info function"""

    def test_might_fail(self):
        """Verify that cli_cmd_info works without mocks"""
        # the `free -m` command is most likely on every system, thus most likely to work
        data = iiq_gather_info.cli_cmd_info('free -m', iiq_gather_info.cli_parsers.memory_to_dict)
        the_json = json.loads(data)
        expected_cmd = 'free -m'

        self.assertEqual(the_json['command'], expected_cmd)

    def test_bad_command(self):
        """Supplying a bad CLI command to `cli_cmd_info` returns valid JSON still"""
        data = iiq_gather_info.cli_cmd_info('sdfwehsxiodaweh', iiq_gather_info.cli_parsers.memory_to_dict)
        the_json = json.loads(data)

        self.assertTrue(the_json['exitcode'] != 0)

    def test_parser_failure(self):
        """A good command, and parser failure still generates valid JSON"""
        data = iiq_gather_info.cli_cmd_info('date', iiq_gather_info.cli_parsers.memory_to_dict)
        the_json = json.loads(data)

        self.assertTrue(the_json['traceback'].startswith('Traceback'))


class TestCliCmdInfoHelpers(unittest.TestCase):
    """A suite of tests for the helpers for obtaining CLI command output"""

    @patch.object(iiq_gather_info, 'cli_cmd_info')
    def test_mount_info(self, fake_cli_cmd_info):
        """iiq_gather_info.mount_info calls the expected command"""
        iiq_gather_info.mount_info()
        args, _ = fake_cli_cmd_info.call_args
        sent_command = args[0]
        expected_command = 'df -P'

        self.assertEqual(sent_command, expected_command)

    @patch.object(iiq_gather_info, 'cli_cmd_info')
    def test_memory_info(self, fake_cli_cmd_info):
        """iiq_gather_info.memory_info calls the expected command"""
        iiq_gather_info.memory_info()
        args, _ = fake_cli_cmd_info.call_args
        sent_command = args[0]
        expected_command = 'free -m'

        self.assertEqual(sent_command, expected_command)

    @patch.object(iiq_gather_info, 'cli_cmd_info')
    def test_ifconfig_info(self, fake_cli_cmd_info):
        """iiq_gather_info.ifconfig_info calls the expected command"""
        iiq_gather_info.ifconfig_info()
        args, _ = fake_cli_cmd_info.call_args
        sent_command = args[0]
        expected_command = 'ifconfig'

        self.assertEqual(sent_command, expected_command)

    @patch.object(iiq_gather_info, 'versions')
    def test_iiq_version_info(self, fake_versions):
        """iiq_gather_info.iiq_version_info calls the expected command"""
        fake_versions.get_iiq_version.return_value = Version(version='1.2.3', name='InsightIQ')
        fake_versions.get_iiqtools_version.return_value = Version(version='1.2.3', name='InsightIQ')
        data = iiq_gather_info.iiq_version_info()
        expected = {'insightiq' : '1.2.3', 'iiqtools' : '1.2.3'}

        self.assertEqual(json.loads(data), expected)

class TestIiqGatheInfoMain(unittest.TestCase):
    """A suite of tests for the iiq_gather_info.main function"""

    @patch.object(iiq_gather_info, 'versions')
    @patch.object(iiq_gather_info.sys, 'stdout')
    @patch.object(__builtin__, 'raw_input')
    @patch.object(iiq_gather_info, 'get_tarfile')
    @patch.object(iiq_gather_info, 'get_logger')
    def test_main(self, fake_get_logger, fake_get_tarfile, fake_raw_input, fake_stdout, fake_versions):
        """iiq_gather_info.main is callable, and returns an integer"""
        fake_versions.get_iiq_version.return_value = Version(version='1.2.3', name='InsightIQ')
        fake_versions.get_iiqtools_version.return_value = Version(version='1.2.3', name='InsightIQ')
        fake_raw_input.return_value = 'yes'
        exit_code = iiq_gather_info.main(['--case-number', '0', '--output-dir', '/tmp'])
        expected = 0

        self.assertEqual(exit_code, expected)

    @patch.object(iiq_gather_info.sys, 'stdout')
    @patch.object(__builtin__, 'raw_input')
    @patch.object(iiq_gather_info, 'get_tarfile')
    @patch.object(iiq_gather_info, 'get_logger')
    def test_not_root_and_no(self, fake_get_logger, fake_get_tarfile, fake_raw_input, fake_stdout):
        """If not ran as root, and you say `no` to the prompt, the script exit code is 1"""
        fake_raw_input.return_value = 'no'
        exit_code = iiq_gather_info.main(['--case-number', '0', '--output-dir', '/tmp'])
        expected = 1

        self.assertEqual(exit_code, expected)


    @patch.object(iiq_gather_info.sys, 'stdout')
    @patch.object(__builtin__, 'raw_input')
    @patch.object(iiq_gather_info, 'get_tarfile')
    @patch.object(iiq_gather_info, 'get_logger')
    def test_ioerror(self, fake_get_logger, fake_get_tarfile, fake_raw_input, fake_stdout):
        """An I/O (or OS) error when creating the tarfile returns a non-zero exit code"""
        fake_raw_input.return_value = 'yes'
        fake_get_tarfile.side_effect = [IOError(13, 'testing', 'some_file')]

        exit_code = iiq_gather_info.main(['--case-number', '0', '--output-dir', '/tmp'])
        expected = 13

        self.assertEqual(exit_code, expected)


if __name__ == '__main__':
    unittest.main()



# -*- coding: UTF-8 -*-
"""
Unit tests for the Version object
"""
import unittest
from mock import patch, MagicMock

from iiqtools.utils import versions


class TestInit(unittest.TestCase):
    """A suite of tests for non-comparison operator specific functionality"""

    def test_init_ok(self):
        """Able to instantiate the Version object"""
        v1 = versions.Version(version='1.2.3', name='foo')
        self.assertTrue(isinstance(v1, versions.Version))

    def test_invalid_version_ints(self):
        """Version only support integer version values"""
        self.assertRaises(ValueError, versions.Version, version='1a.2', name='foo')

    def test_invalid_version_fields(self):
        """Version requires at least Major.Minor for verisons"""
        self.assertRaises(ValueError, versions.Version, version='1234', name='foo')

    def test_invalid_version_value(self):
        """Version only accepts strings for version param"""
        self.assertRaises(TypeError, versions.Version, version=1, name='foo')

    def test_name_property_ro(self):
        """Version.name is read only"""
        v1 = versions.Version(version='1.2.3', name='foo')
        try:
            v1.name = 'bar'
        except AttributeError:
            passed = True
        else:
            passed = False

        self.assertTrue(passed)

    def test_semver_property_ro(self):
        """Version.semver is read only"""
        v1 = versions.Version(version='1.2.3', name='foo')
        try:
            v1.semver = 'bar'
        except AttributeError:
            passed = True
        else:
            passed = False

        self.assertTrue(passed)

    def test_version_property_ro(self):
        """Version.version is read only"""
        v1 = versions.Version(version='1.2.3', name='foo')
        try:
            v1.version = '3.4'
        except AttributeError:
            passed = True
        else:
            passed = False

        self.assertTrue(passed)

    def test_major_property_ro(self):
        """Version.major is read only"""
        v1 = versions.Version(version='1.2.3', name='foo')
        try:
            v1.major = 12
        except AttributeError:
            passed = True
        else:
            passed = False

        self.assertTrue(passed)

    def test_minor_property_ro(self):
        """Version.minor is read only"""
        v1 = versions.Version(version='1.2.3', name='foo')
        try:
            v1.minor = 33
        except AttributeError:
            passed = True
        else:
            passed = False

        self.assertTrue(passed)

    def test_patch_property_ro(self):
        """Version.patch is read only"""
        v1 = versions.Version(version='1.2.3', name='foo')
        try:
            v1.patch = 234
        except AttributeError:
            passed = True
        else:
            passed = False

        self.assertTrue(passed)

    def test_build_property_ro(self):
        """Version.build is read only"""
        v1 = versions.Version(version='1.2.3', name='foo')
        try:
            v1.build = 9001
        except AttributeError:
            passed = True
        else:
            passed = False

        self.assertTrue(passed)

    def test_name_property(self):
        """Version.name is the expected value"""
        v1 = versions.Version(version='1.2.3', name='foo')
        expected = 'foo'

        self.assertEqual(v1.name, expected)

    def test_semver_property(self):
        """Version.semver is read only"""
        v1 = versions.Version(version='1.2.3', name='foo')
        expected = ('major', 'minor', 'patch', 'build')

        self.assertEqual(v1.semver, expected)

    def test_version_property(self):
        """Version.version is read only"""
        v1 = versions.Version(version='1.2.3', name='foo')
        expected = '1.2.3'

        self.assertEqual(v1.version, expected)

    def test_major_property(self):
        """Version.major is read only"""
        v1 = versions.Version(version='1.2.3', name='foo')
        expected = 1

        self.assertEqual(v1.major, expected)

    def test_minor_property(self):
        """Version.minor is read only"""
        v1 = versions.Version(version='1.2.3', name='foo')
        expected = 2

        self.assertEqual(v1.minor, expected)

    def test_patch_property(self):
        """Version.patch is read only"""
        v1 = versions.Version(version='1.2.3', name='foo')
        expected = 3

        self.assertEqual(v1.patch, expected)

    def test_build_property(self):
        """Version.build is read only"""
        v1 = versions.Version(version='1.2.3.4', name='foo')
        expected = 4

        self.assertEqual(v1.build, expected)

    def test_undefined_semver(self):
        """Version - properties that are undefined are ``None``"""
        v1 = versions.Version(version='1.2.3', name='foo')
        expected = None

        self.assertEqual(v1.build, expected)

    def test_hash(self):
        """Version objects are hashable, i.e. can be added to a dictionary"""
        v1 = versions.Version(version='1.2.3', name='foo')
        my_dict = { v1 : 'foo'}

        self.assertTrue(isinstance(my_dict, dict))

    def test_get_other_typeerror(self):
        """Version._get_other supports only strings or Version as param"""
        v = versions.Version(name='foo', version='1.2.3')
        self.assertRaises(TypeError, v._get_other, 3.4)

    def test_get_other_typeerror_2(self):
        """Version._get_other - Invalid version strings raises a TypeError"""
        v = versions.Version(name='foo', version='1.2.3')
        self.assertRaises(TypeError, v._get_other, '1')


class TestEquals(unittest.TestCase):
    """A suite of tests that compare equality, ``==``"""

    def test_a(self):
        """Version - identical versions values are equal, regardless of name"""
        v1 = versions.Version(version='1.2.3', name='foo')
        v2 = versions.Version(version='1.2.3', name='bar')

        self.assertTrue(v1 == v2)
        self.assertTrue(v2 == v1)

    def test_b(self):
        """Version - versions that are longer are not equal, regardless of name"""
        v1 = versions.Version(version='1.2.3', name='foo')
        v2 = versions.Version(version='1.2', name='bar')

        self.assertFalse(v1 == v2)
        self.assertFalse(v2 == v1)

    def test_c(self):
        """Version - different versions are not equal, regardless of name"""
        v1 = versions.Version(version='1.2.3', name='foo')
        v2 = versions.Version(version='2.2.3', name='bar')

        self.assertFalse(v1 == v2)
        self.assertFalse(v2 == v1)

    def test_d(self):
        """Version - the equals operator works with strings"""
        v1 = versions.Version(version='1.2.3', name='foo')

        self.assertTrue(v1 == '1.2.3')


class TestNotEquals(unittest.TestCase):
    """A suite of test that compare not equivalence, ``!=``"""

    def test_a(self):
        """Version - different versions are not equal, regardless of name"""
        v1 = versions.Version(version='1.2.3', name='foo')
        v2 = versions.Version(version='2.2.3', name='bar')

        self.assertTrue(v1 != v2)
        self.assertTrue(v2 != v1)

    def test_b(self):
        """Version - versions that are of different lengths are not equal, regardless of name"""
        v1 = versions.Version(version='1.2.3', name='foo')
        v2 = versions.Version(version='1.2', name='bar')

        self.assertTrue(v1 != v2)
        self.assertTrue(v2 != v1)


    def test_c(self):
        """Version - the same versions are not not-equal, regardless of name"""
        v1 = versions.Version(version='1.2.3', name='foo')
        v2 = versions.Version(version='1.2.3', name='bar')

        self.assertFalse(v1 != v2)
        self.assertFalse(v2 != v1)

    def test_d(self):
        """Version the not equals operator works with strings"""
        v1 = versions.Version(version='1.2.3', name='foo')

        self.assertTrue(v1 != '3.4')


class TestLessThan(unittest.TestCase):
    """A suite of test that compare less than, ``<``"""

    def test_a(self):
        """Version - equal versions are not less than"""
        v1 = versions.Version(version='1.2.3', name='foo')
        v2 = versions.Version(version='1.2.3', name='bar')

        self.assertFalse(v1 < v2)
        self.assertFalse(v2 < v1)

    def test_b(self):
        """Version - Less specific versions that are otherwise identical are less than"""
        v1 = versions.Version(version='1.2.3', name='foo')
        v2 = versions.Version(version='1.2', name='bar')

        self.assertFalse(v1 < v2)
        self.assertTrue(v2 < v1)

    def test_c(self):
        """Version - ``None`` is less than zero"""
        v1 = versions.Version(version='1.2.0', name='foo')
        v2 = versions.Version(version='1.2', name='bar')

        self.assertFalse(v1 < v2)
        self.assertTrue(v2 < v1)

    def test_d(self):
        """Version - Larger versions are not less than"""
        v1 = versions.Version(version='1.2.1', name='foo')
        v2 = versions.Version(version='1.2.2', name='bar')

        self.assertTrue(v1 < v2)
        self.assertFalse(v2 < v1)

    def test_e(self):
        """Version - The less than operator works with strings"""
        v1 = versions.Version(version='1.2.1', name='foo')

        self.assertTrue(v1 < '3.4.5.6')


class TestGreaterThan(unittest.TestCase):
    """A suite of test that compare less than, ``>``"""

    def test_a(self):
        """Version - equal versions are not greater than"""
        v1 = versions.Version(version='1.2.3', name='foo')
        v2 = versions.Version(version='1.2.3', name='bar')

        self.assertFalse(v1 > v2)
        self.assertFalse(v2 > v1)

    def test_b(self):
        """Version - Less specific versions that are otherwise identical are less than"""
        v1 = versions.Version(version='1.2.3', name='foo')
        v2 = versions.Version(version='1.2', name='bar')

        self.assertFalse(v1 < v2)
        self.assertTrue(v2 < v1)

    def test_c(self):
        """Version - zero is greater than ``None``"""
        v1 = versions.Version(version='1.2.0', name='foo')
        v2 = versions.Version(version='1.2', name='bar')

        self.assertTrue(v1 > v2)
        self.assertFalse(v2 > v1)

    def test_d(self):
        """Version - Larger versions are greater than"""
        v1 = versions.Version(version='1.2.1', name='foo')
        v2 = versions.Version(version='1.2.2', name='bar')

        self.assertFalse(v1 > v2)
        self.assertTrue(v2 > v1)

    def test_e(self):
        """Version - The greater than operator works with strings"""
        v1 = versions.Version(version='3.2.1', name='foo')

        self.assertTrue(v1 > '2.3')


class TestLessThanEqualTo(unittest.TestCase):
    """A suite of test that compare less than, ``<=``"""

    def test_a(self):
        """Version - Larger version are not less than or equal to"""
        v1 = versions.Version(version='1.2.1', name='foo')
        v2 = versions.Version(version='1.2.2', name='bar')

        self.assertTrue(v1 <= v2)
        self.assertFalse(v2 <= v1)

    def test_b(self):
        """Version - Equal versions are less than or equal to"""
        v1 = versions.Version(version='1.2.1', name='foo')
        v2 = versions.Version(version='1.2.1', name='bar')

        self.assertTrue(v1 <= v2)
        self.assertTrue(v2 <= v1)

    def test_c(self):
        """Version - Less specific versions are less than or equal"""
        v1 = versions.Version(version='1.2', name='foo')
        v2 = versions.Version(version='1.2.1', name='bar')

        self.assertTrue(v1 <= v2)
        self.assertFalse(v2 <= v1)

    def test_d(self):
        """Version - The less than or equal to operator works with strings"""
        v1 = versions.Version(version='1.2', name='foo')

        self.assertTrue(v1 <= '1.2')


class TestGreaterThanEqualTo(unittest.TestCase):
    """A suite of test that compare less than, ``>=``"""

    def test_a(self):
        """Version - Larger version are not less than or equal to"""
        v1 = versions.Version(version='1.2.1', name='foo')
        v2 = versions.Version(version='1.2.2', name='bar')

        self.assertFalse(v1 >= v2)
        self.assertTrue(v2 >= v1)

    def test_b(self):
        """Version - Equal versions are less than or equal to"""
        v1 = versions.Version(version='1.2.1', name='foo')
        v2 = versions.Version(version='1.2.1', name='bar')

        self.assertTrue(v1 >= v2)
        self.assertTrue(v2 >= v1)

    def test_c(self):
        """Version - Less specific versions are less than or equal"""
        v1 = versions.Version(version='1.2', name='foo')
        v2 = versions.Version(version='1.2.1', name='bar')

        self.assertFalse(v1 >= v2)
        self.assertTrue(v2 >= v1)

    def test_d(self):
        """Version - Greather than or equal to operator supports strings"""
        v1 = versions.Version(version='1.2', name='foo')

        self.assertTrue(v1 >= '1.0')


class TestGetVersions(unittest.TestCase):
    """A suite of tests for the ``get_iiq_version`` and ``get_iiqtools_version`` functions"""

    class FakeDistVersion(object):
        """For replacing what ``get_distribution`` returns in testing"""
        def __init__(self, version):
            self.version = version

    @patch.object(versions, 'get_distribution')
    def test_get_iiq_version(self, fake_get_distribution):
        """None is returned if InsightIQ is not installed"""
        fake_get_distribution.side_effect = [versions.DistributionNotFound()]

        v = versions.get_iiq_version()

        self.assertTrue(v is None)

    @patch.object(versions, 'get_distribution')
    def test_get_iiqtools_version(self, fake_get_distribution):
        """None is returned if IIQTools is not installed"""
        fake_get_distribution.side_effect = [versions.DistributionNotFound()]

        v = versions.get_iiqtools_version()

        self.assertTrue(v is None)

    @patch.object(versions, 'get_distribution')
    def test_get_iiq_version_ok(self, fake_get_distribution):
        """Version is returned if InsightIQ is  installed"""
        fake_get_distribution.return_value = self.FakeDistVersion('3.3.4')

        v = versions.get_iiq_version()

        self.assertTrue(isinstance(v, versions.Version))

    @patch.object(versions, 'get_distribution')
    def test_get_iiqtools_version_ok(self, fake_get_distribution):
        """Version is returned if IIQTools is installed"""
        fake_get_distribution.return_value = self.FakeDistVersion('1.2.3')

        v = versions.get_iiqtools_version()

        self.assertTrue(isinstance(v, versions.Version))

class TestPatchInfo(unittest.TestCase):
    """TODO"""

    def test_patch_info(self):
        """The PatchInfo API accepts the expected params"""
        patch_info = versions.PatchInfo(iiq_dir='/some/path',
                                         patches_dir='/another/path',
                                         is_installed=False,
                                         specific_patch='this_patch',
                                         readme='data from readme',
                                         all_patches=('patch1', 'patch2'))
        self.assertTrue(isinstance(patch_info, versions._PatchInfo))

    def test_get_patch_info_returns(self):
        """versions.get_patch_info always returns a PatchInfo object"""
        # This test assumes IIQ isn't installed, thus the pile of errors that'll
        # occur shouldn't prevent us from getting a PatchInfo object
        fake_log = MagicMock()
        patch_info = versions.get_patch_info('bogus-patch.tgz', fake_log)

        self.assertTrue(isinstance(patch_info, versions._PatchInfo))
        self.assertEqual(patch_info.iiq_dir, '')
















if __name__ == '__main__':
    unittest.main()



# -*- coding: UTF-8 -*-
"""
Utilities for interacting with the InsightIQ database
"""
from collections import namedtuple

import psycopg2

from iiqtools.exceptions import DatabaseError


_Column = namedtuple('Column', 'name type')

# This allows us to define the docstring for out namedtuple
class Column(_Column):
    """A database column consiting of the column's name, and it's type

    :Type: namedtuple

    :param name: The column's name

    :param type: The database type for the given column (i.e. int, float, double)
    """
    pass

# This database abstraction intentionally does not comply with PEP 249
# https://www.python.org/dev/peps/pep-0249/
# Why? It's a bit much if you're not an experienced developer or you only
# need to perform simple interactions with a database. The point of this
# abstraction to have a simple, easy to use object for poking the IIQ database
class Database(object):
    """Simplifies communication with the database.

    The goal of this object is to make basic interactions with the database
    simpler than directly using the psycopg2 library. It does this by reducing
    the number of API methods, providing handy built-in methods for common needs
    (like listing tables of a database), auto-commit of transactions, and
    auto-rollback of bad SQL transactions.
    This object is not indented for power users, or long lived processes (like
    the InsightIQ application); it's designed for shorter lived "scripts".

    :param user: The username when connection to the databse
    :type user: String, default postgres

    :param dbname: The specific database to connection to. InsightIQ utilizes
                   a different database for every monitored cluster, plus one
                   generic database for the application (named "insightiq").
    :type dbname: String, default insightiq
    """
    def __init__(self, user='postgres', dbname='insightiq'):
        self._connection = psycopg2.connect(user=user, dbname=dbname)
        self._cursor = self._connection.cursor()
        if dbname == 'insightiq':
            self.execute("SET search_path to admin,iiq,public;")

    def __enter__(self):
        """Enables use of the ``with`` statement to auto close database connection
        https://docs.python.org/2.7/reference/datamodel.html#with-statement-context-managers

        Example::

          with Database() as db:
              print db.cluster_databases
        """
        return self

    def __exit__(self, exc_type, exc_value, the_traceback):
        self._connection.close()

    def execute(self, sql, params=None):
        """Run a single SQL command

        :Returns: Generator

        :param sql: **Required** The SQL syntax to execute
        :type sql: String

        :param params: The values to use in a parameterized SQL query
        :type params: Iterable

        This method is implemented as a Python Generator:
        https://wiki.python.org/moin/Generators
        This means you are suppose to iterate over the results::

            db = Database()
            for row in db.execute("select * from some_table;"):
                print row

        If you want all the rows as a single thing, just use ``list``::

            db = Database()
            data = list(db.execute("select * from some_table;")

        But **WARNING** that might cause your program to run out of memory and crash!
        That reason is why this method is a generator by default ;)

        To perform a parameterized query (i.e. avoid SQL injection), provided
        the parameters as an iterable::

            db = Database()
            # passing in "foo_column" alone would try and string format every
            # character of "foo_column" into your SQL statement.
            # Instead, make "foo_column" a tuple by wrapping it like ("foo_column",)
            # Note: the trailing comma is required.
            data = list(db.execute("select %s from some_table", ("foo_column",)))
        """
        return self._query(sql, params=params, many=False)

    def executemany(self, sql, params):
        """Run the SQL for every iteration of the supplied params

        This method behaves exactly like `execute`, except that it can perform
        multiple SQL commands in a single transaction. The point of this method
        is so you can retain Atomicity when you must execute the same SQL with
        different parameters. This method isn't intended to be faster than
        looping over the normal `execute` method with the different parameters.

        :Returns: Generator

        :param sql: **Required** The SQL syntax to execute
        :type sql: String

        :param params: **Required** The parameterized values to iterate
        :type params: Iterable
        """
        return self._query(sql, params=params, many=True)

    def _query(self, sql, params=None, many=False):
        """Internal method for running SQL commands

        The code difference between execute, and executemany is just the method
        we call on the cursor object.

        :Returns: Generator

        :param sql: **Required** The SQL syntax to execute
        :type sql: String

        :param params: The values to use in a parameterized SQL query
        :type params: Iterable

        :param many: Set to True to call `executemany`
        :type many: Boolean, default is False
        """
        if many:
            call = getattr(self._cursor, 'executemany')
        else:
            call = getattr(self._cursor, 'execute')
        try:
            call(sql, params)
            self._connection.commit()
        except psycopg2.Error as doh:
            # All psycopg2 Exceptions are subclassed from psycopg2.Error
            self._connection.rollback()
            raise DatabaseError(message=doh.pgerror, pgcode=doh.pgcode)
        else:
            data = self._cursor.fetchone()
            while data:
                yield data
                data = self._cursor.fetchone()

    @property
    def isolation_level(self):
        """Set the isolation level of your connnection to the database"""
        # To drop tables, you'll have to set isolation_level to 0 (zero)
        # https://www.postgresql.org/docs/current/static/transaction-iso.html
        return self._connection.isolation_level

    @isolation_level.setter
    def isolation_level(self, value):
        # https://www.postgresql.org/docs/current/static/transaction-iso.html
        self._connection.set_isolation_level(value)

    def tables(self):
        """Obtain a list of all the tables for the database you're connected to

        :Returns: List
        """
        sql = """SELECT relname FROM pg_class WHERE relkind='r' AND relname !~ '^(pg_|sql_)';"""
        # data looks like [('sometable1',), ('sometable2',)]
        # drop all the tuples, so we just return a list of strings
        return [x[0] for x in self.execute(sql)]

    def cluster_databases(self):
        """Obtain a list of all the cluster databases

        :Returns: List
        """
        sql = """SELECT datname from pg_database;"""
        dbs = list(self.execute(sql))
        ignore = ('template0', 'template1', 'postgres', 'insightiq')
        return [x[0] for x in dbs if x[0] not in ignore]

    def table_schema(self, table):
        """Given a table, return the schema for that table

        :Returns: Tuple of namedtuples -> (Column(name, type), Column(name, type))

        :param table: **Required** The table to obtain the primary key from
        :type table: String
        """
        sql = """SELECT column_name, data_type
                 FROM INFORMATION_SCHEMA.COLUMNS WHERE table_name = %s;"""
        return tuple([Column(x[0], x[1]) for x in self.execute(sql, (table,))])

    def primary_key(self, table):
        """Given a table, return the primary key

        .. note::

            If you supply a timeseries table that DOES NOT have an EPOC timestamp
            in the name, you will get zero results. For timeseries tables,
            supply a table that contains the EPOC timestamps to see the primary key.

        :Returns: Tuple of namedtuples -> (Column(name, type), Column(name, type))

        :param table: **Required** The table to obtain the primary key from
        :type table: String
        """
        sql = """SELECT a.attname, format_type(a.atttypid, a.atttypmod) AS data_type
                 FROM   pg_index i
                 JOIN   pg_attribute a ON a.attrelid = i.indrelid
                                        AND a.attnum = ANY(i.indkey)
                 WHERE  i.indrelid = %s::regclass
                 AND    i.indisprimary;"""
        return tuple([Column(x[0], x[1]) for x in self.execute(sql, (table,))])

    def close(self):
        """Disconnect from the database"""
        self._connection.close()



# -*- coding: UTF-8 -*-
"""
Unit tests for the iiqtools.utils.logger module
"""
import logging
import unittest

from iiqtools.utils.logger import get_logger


class TestGetLogger(unittest.TestCase):
    """A suite of test cases for the iiqtools.utils.logger module"""

    def setUp(self):
        """Runs before every test case"""
        # clear out all logger objects
        logging.Logger.manager.loggerDict = {}

    def test_happy_path(self):
        """You'll get a logger if you only supply a valid path to log to"""
        log = get_logger(log_path='/dev/null')
        self.assertTrue(isinstance(log, logging.Logger))

    def test_bad_path(self):
        """AssertionError is raised for an invalid file system path"""
        self.assertRaises(AssertionError, get_logger, log_path='not a path')

    def test_bad_stream_lvl(self):
        """AssertionError is raised if the param stream_lvl is not a valid integer"""
        self.assertRaises(AssertionError, get_logger, log_path='/dev/null', stream_lvl=9001)

    def test_bad_file_lvl(self):
        """AssertionError is raised if the param file_lvl is not a valid integer"""
        self.assertRaises(AssertionError, get_logger, log_path='/dev/null', file_lvl=9001)

    def test_log_level_no_stream(self):
        """The log level will be the value of file_lvl when stream_lvl is zero"""
        file_lvl = 30
        log = get_logger(log_path='/dev/null', file_lvl=file_lvl)
        self.assertEqual(log.level, file_lvl)

    def test_log_level_with_stream(self):
        """The log level will be the lower value between stream_lvl and file_lvl"""
        stream_lvl = 20
        file_lvl = 40
        log =  get_logger(log_path='/dev/null', file_lvl=file_lvl, stream_lvl=stream_lvl)
        self.assertEqual(log.level, stream_lvl)

    def test_calling_get_logger_mutiple_times(self):
        """Asking for the same logger multiple times returns the same object"""
        log1 = get_logger(log_path='/dev/null')
        log2 = get_logger(log_path='/dev/null')
        self.assertTrue(log1 is log2)

    def test_get_logger_mutiple_times_no_extra_handlers(self):
        """Asking for the same logger multiple times returns the same object
        without adding more handelers
        """
        log1 = get_logger(log_path='/dev/null')
        log1_handler_count = len(log1.handlers)
        log2 = get_logger(log_path='/dev/null')
        log2_handler_count = len(log2.handlers)
        self.assertTrue(log1_handler_count is log2_handler_count)


if __name__ == '__main__':
    unittest.main()



# -*- coding: UTF-8 -*-
"""
A suite of tests for the iiq_version module
"""
import __builtin__
import unittest
from mock import patch

from iiqtools import iiq_version


class TestParseCli(unittest.TestCase):
    """A suite of tests for the parse_cli function in iiqtools.iiq_version"""

    def test_basic(self):
        """Supplying no args returns an argparse.Namespace object"""
        args = iiq_version.parse_cli([])
        self.assertTrue(isinstance(args, iiq_version.argparse.Namespace))

    # Mocking away stderr to avoid spam in output while running tests
    @patch.object(iiq_version.argparse._sys, 'stderr')
    def test_takes_no_args(self, fake_stderr):
        """iiq_version doesn't accept any CLI args (besides -h/--help)"""
        self.assertRaises(SystemExit, iiq_version.parse_cli, ['-f'])


class TestMain(unittest.TestCase):
    """A suite of tests for the ``main`` function of iiqtools.iiq_version"""

    @patch.object(__builtin__, 'print')
    def test_foo(self, fake_print):
        """InsightIQ and IIQTools are printed to the console"""
        iiq_version.main([])
        calls = fake_print.call_args_list

        self.assertEqual(len(calls), 2)

if __name__ == '__main__':
    unittest.main()



# -*- coding: UTF-8 -*-
"""
Unit tests for the iiqtools.exceptions module
"""

import unittest

from iiqtools import exceptions


class TestExceptions(unittest.TestCase):
    """A suite of test cases for the iiqtools.exceptions module"""

    def test_cli_error(self):
        """CliError is a valid exception"""
        exception = exceptions.CliError(command='foo', stdout='woot', stderr='boo', exit_code=1)
        self.assertTrue(isinstance(exception, exceptions.CliError))

    def test_database_error(self):
        """DatabaseError is a valid exception"""
        exception = exceptions.DatabaseError(message='oops', pgcode='asdf')
        self.assertTrue(isinstance(exception, exceptions.DatabaseError))


if __name__ == '__main__':
    unittest.main()
